-- http://blog.mathieu-leplatre.info/use-postgis-topologies-to-clean-up-road-networks.html
--SELECT DropTopology('ch13_topology');
-- DROP SCHEMA IF EXISTS ch13 CASCADE;
-- SELECT setval('topology.topology_id_seq', 1, true);
--
-- <start id="code_create_ch13_staging" /> --
CREATE SCHEMA ch13_staging;
CREATE SCHEMA ch13;
-- <end id="code_create_ch13_staging" /> --


-- <start id="code_create_ch13_topology" /> --
SELECT CreateTopology('ch13_topology', 32610, 0.05);
-- <end id="code_create_ch13_topology" /> --

-- <start id="code_create_ch13_admin_boundaries_line" /> --
SELECT gid, TopoGeo_AddLineString('ch13_topology'
  , ST_Transform(geom,32610)) As edge_id -- <co id="co_code_create_ch13_admin_boundaries_line_1" />
FROM (SELECT gid, (ST_Dump(geom)).geom  -- <co id="co_code_create_ch13_admin_boundaries_line_2" />
 FROM ch13_staging.cityboundary
) As f;
-- <end id="code_create_ch13_admin_boundaries_line" /> --
-- [1] create edges
-- [2] expand multiline to individual line

-- <start id="code_count_faces_edges_nodes_relations" /> --
SELECT 'faces' As type, count(*) As num
  FROM ch13_topology.face
UNION ALL
SELECT 'edges' As type, count(*) As num
  FROM ch13_topology.edge
UNION ALL
SELECT 'nodes' As type, count(*) As num
  FROM ch13_topology.node
UNION ALL
SELECT 'relations' As type, count(*) As num
FROM ch13_topology.relation;
-- <end id="code_count_faces_edges_nodes_relations" /> --

-- <start id="code_create_topogeo_addpolygon_municipal" /> --
SELECT gid
 , TopoGeo_AddPolygon('ch13_topology', 
   ST_Transform(geom,32610), 0.05) As face_id
FROM (SELECT gid, (ST_Dump(geom)).geom 
   FROM ch13_staging.neighbourhoods) As f;
-- <end id="code_create_topogeo_addpolygon_municipal" /> --

-- <start id="code_build_topogeometry_columns_prim_child" /> --
CREATE TABLE ch13.neighbourhoods(feat_name varchar(50) primary key);
SELECT AddTopoGeometryColumn('ch13_topology', 'ch13',    -- <co id="co_code_build_topogeometry_columns_prim_child_1" /> --
  'neighbourhoods', 'topo', 'MULTIPOLYGON');
  
CREATE TABLE ch13.cities(feat_name varchar(150) primary key);
SELECT AddTopoGeometryColumn('ch13_topology', 'ch13',   
  'cities', 'topo', 'MULTIPOLYGON', 1);  -- <co id="co_code_build_topogeometry_columns_prim_child_2" /> --
-- <end id="code_build_topogeometry_columns_prim_child" /> --

-- <start id="code_totopogeom_nei" /> --
INSERT INTO ch13.neighbourhoods(feat_name, topo)
SELECT  neighbourh,  
  toTopoGeom(ST_Transform(geom,32610), 'ch13_topology', 1, 0.05)
FROM ch13_staging.neighbourhoods;
-- <end id="code_totopogeom_nei" /> --

-- <start id="code_createtopogeom_city" /> --
INSERT INTO ch13.cities(feat_name, topo)
SELECT 'Victoria',    
  CreateTopoGeom('ch13_topology', 3,  -- <co id="co_code_createtopogeom_city_1" /> --
       2,    -- <co id="co_code_createtopogeom_city_2" /> --
  (SELECT TopoElementArray_Agg(
     ARRAY[(n.topo).id, (n.topo).layer_id]) -- <co id="co_code_createtopogeom_city_3" /> --
      FROM ch13.neighbourhoods AS n) );
-- <end id="code_createtopogeom_city" /> --

-- [1] areal
-- [2] layer
-- [3] elements

-- <start id="code_city_asgeom" /> --
SELECT topo::geometry 
FROM ch13.cities
WHERe feat_name = 'Victoria';
-- <end id="code_city_asgeom" /> --


-- <start id="code_getfacegeom_city" /> --
SELECT face_id,
 ST_GetFaceGeometry('ch13_topology', face_id)
FROM (SELECT (GetTopoGeomElements(topo))[1]  As face_id 
    FROM ch13.cities
WHERe feat_name = 'Victoria') As f;
-- <end id="code_getfacegeom_city" /> --

--<start id="code_count_faces_neighbourhood" /> --
SELECT feat_name, count(face_id) As num_faces, 
 MIN(
  ST_Area(ST_GetFaceGeometry('ch13_topology', face_id)) 
  )::numeric(10,2) As min_area,
 MAX(
  ST_Area(ST_GetFaceGeometry('ch13_topology', face_id)) 
  )::numeric(10,2) As max_area
FROM (SELECT feat_name, (GetTopoGeomElements(topo))[1]  As face_id 
    FROM ch13.neighbourhoods) As f
GROUP BY feat_name 
HAVING count(face_id) > 1
ORDER BY count(face_id) DESC;
-- <end id="code_count_faces_neighbourhood" /> --

--<start id="code_small_face_pockets" /> --
SELECT face_id, feat_name, (ST_GetFaceEdges('ch13_topology', face_id)).edge, 
 ST_Area(ST_GetFaceGeometry('ch13_topology', face_id)) As area
FROM (SELECT feat_name, (GetTopoGeomElements(topo))[1]  As face_id 
    FROM ch13.neighbourhoods ) As f
    WHERE ST_Area(ST_GetFaceGeometry('ch13_topology', face_id)) < 100
    ORDER BY feat_name, area;
--<end id="code_small_face_pockets" /> --

-- remove face edges --
-- <start id="code_remove_small_pockets" /> --
DO language plpgsql $$DECLARE r record; var_face integer;
BEGIN
  FOR r IN ( 
   SELECT DISTINCT abs(
    (ST_GetFaceEdges('ch13_topology', face_id)).edge ) As edge 
FROM (SELECT feat_name, (GetTopoGeomElements(topo))[1]  As face_id 
    FROM ch13.neighbourhoods ) As f
    WHERE ST_Area(ST_GetFaceGeometry('ch13_topology', face_id)) < 55000
     ) LOOP
    BEGIN
      var_face := ST_RemEdgeNewFace('ch13_topology', r.edge);
    EXCEPTION
      WHEN OTHERS THEN
        RAISE WARNING 'Failed remove edge: %, %', r.edge, SQLERRM;
    END;
  END LOOP;
END$$;
-- <end id="code_remove_small_pockets" /> -

-- <start id="code_determine_faces_shared" /> --
SELECT (GetTopoGeomElements(topo))[1] As face_id -- <co id="co_code_determine_faces_shared_1" /> --
 FROM ch13.neighbourhoods As n
 WHERE feat_name = 'North Park' 
INTERSECT    -- <co id="co_code_determine_faces_shared_2" /> --
SELECT (GetTopoGeomElements(topo))[1] As face_id  -- <co id="co_code_determine_faces_shared_3" /> --
 FROM ch13.neighbourhoods As n
 WHERE feat_name = 'Burnside' ;
-- <end id="code_determine_faces_shared" /> --

-- <start id="code_removing_faces_topogeom" /> --
DELETE FROM ch13_topology.relation AS r  -- <co id="co_code_removing_faces_topogeom_1" /> --
WHERE EXISTS (SELECT topo 	-- <co id="co_code_removing_faces_topogeom_2" /> --
  FROM ch13.neighbourhoods As n 
  WHERE feat_name = 'North Park'
   AND (topo).id = r.topogeo_id 
   AND r.element_id = 23 AND r.element_type = 3 );

DELETE FROM ch13_topology.relation AS r    -- <co id="co_code_removing_faces_topogeom_3" /> --
WHERE EXISTS (SELECT topo 
  FROM ch13.neighbourhoods As n 
  WHERE feat_name = 'Burnside'
   AND (topo).id = r.topogeo_id 
   AND r.element_id = 31 AND r.element_type = 3 );
-- <end id="code_removing_faces_topogeom" /> --
-- [1] remove relation 
-- [2] corresponding to face 23 associated with North Park topo
-- [3] repeat face 31 Burnside

-- <start id="code_addtopogeomcolumn_streets" /> --
CREATE TABLE ch13.streets(gid integer primary key,  -- <co id="co_code_addtopogeomcolumn_streets_1" /> --
 feat_name varchar(50), access varchar(20), 
 rd_class varchar(20), max_speed numeric(10,2) );
  
SELECT AddTopoGeometryColumn('ch13_topology', 'ch13', -- <co id="co_code_addtopogeomcolumn_streets_2" /> -- 
  'streets', 'topo', 'MULTILINESTRING');
  
CREATE TABLE ch13.log_street_failures   -- <co id="co_code_addtopogeomcolumn_streets_3" /> -- 
(gid integer primary key, error text); 
-- <end id="code_addtopogeomcolumn_streets" /> --  
-- [1] streets
-- [2] add topo
-- [3] log errors


-- 553119ms 
-- <start id="code_totopogeom_streets" /> --
CREATE FUNCTION ch13.load_streets() RETURNS void AS
  $$
  DECLARE r record;
  BEGIN
    FOR r IN SELECT * FROM ch13_staging.streetcentrelines -- <co id="co_code_totopogeom_streets_1" /> --
         ORDER BY gid LIMIT 500 
      OFFSET (SELECT MAX(gid) from ch13.streets) LOOP  
      BEGIN
  INSERT INTO ch13.streets(gid, feat_name, access,  -- <co id="co_code_totopogeom_streets_2" /> --
    rd_class, max_speed, topo)
  SELECT r.gid, r.streetname, r.access, 
    r.rd_class, r.max_speed::numeric,
    toTopoGeom(
     ST_Transform(ST_Force2D(r.geom),32610), -- <co id="co_code_totopogeom_streets_3" /> --
      'ch13_topology', 3, 0.05); -- <co id="co_code_totopogeom_streets_4" /> --
      EXCEPTION
        WHEN OTHERS THEN
          INSERT INTO ch13.log_street_failures(gid, error) -- <co id="co_code_totopogeom_streets_5" /> --
            VALUES (r.gid, SQLERRM);
          RAISE WARNING 'Loading of record % failed: %',
            r.gid, SQLERRM;
      END;
    END LOOP;
  END$$
  language plpgsql;
-- <end id="code_totopogeom_streets" /> --
-- 1 limit
-- 2 insert
-- 3 force2d transform
-- 4 streets layer 0.05 tol
-- 5 log errors

-- <start id="code_pgscript_loop_street_insert" /> --
DECLARE @I;  -- <co id="co_pgscript_loop_street_insert_1" /> --
SET @I = 0;

WHILE @I < 10  -- <co id="co_pgscript_loop_street_insert_2" /> --
BEGIN
  SELECT ch13.load_streets(); -- <co id="co_pgscript_loop_street_insert_3" /> --
  SET @I = @I + 1; -- <co id="co_pgscript_loop_street_insert_4" /> --
  PRINT @I;
END
-- <end id="code_pgscript_loop_street_insert" /> --
-- iterator
-- run 10
-- increment

-- <start id="code_geom_simplification" /> --
SELECT feat_name, ST_Simplify(topo::geometry,150) As geom_simp 
 FROM ch13.neighbourhoods;
-- <end id="code_geom_simplification" /> -- 

-- <start id="code_topo_simplification" /> --
SELECT feat_name, ST_Simplify(topo,150) As topo_simp 
 FROM ch13.neighbourhoods;
-- <end id="code_topo_simplification" /> -- 

-- <start id="code_validatetopology_victoria" /> --
SELECT ValidateTopology('ch13_topology');
-- <end id="code_validatetopology_victoria" /> --

-- <start id="code_topologysummary_victoria" /> --
SELECT TopologySummary('ch13_topology');
-- <end id="code_topologysummary_victoria" /> --

-- NOT USED --
-- <start id="code_totopogeom_new_parcels" /> --
CREATE TABLE ch13.log_parcel_failures(gid integer primary key,
  error text);
DO $$DECLARE r record;
BEGIN
  FOR r IN SELECT * FROM ch13_staging.parcels ORDER BY gid LIMIT 1000 
    OFFSET (SELECT MAX(gid) from ch13.parcels) LOOP
    BEGIN
INSERT INTO ch13.parcels(gid, folio, pid, feat_name, topo)
SELECT r.gid, r.folio, r.pid, r.owneraddre || COALESCE(' ' || r.unit,''),    -- <co id="co_code_totopogeom_new_1" /> --
  toTopoGeom(ST_Force2D(ST_Transform(r.geom,32610)), 'ch13_topology', 2);
    EXCEPTION
      WHEN OTHERS THEN
        INSERT INTO ch13.log_parcel_failures(gid, error) 
          VALUES (r.gid, SQLERRM);
        RAISE WARNING 'Loading of record % failed: %', r.gid, SQLERRM;
    END;
  END LOOP;
END$$;
-- <end id="code_totopogeom_new_parcels" /> --


